/* Miscellaneous functions

   OperStats - Copyright (c) 2000-2002 Darcy Grexton
   Contact: skold@habber.net, skold @ HabberNet, ICQ #6285178

   This program is free but copyrighted software; see the file COPYING
   for details.
 */

#include "../inc/operstats.h"

static FILE *log_file;
static FILE *debug_file;

/* Open the logfile */
void open_log ()
{
    if (log_file)
	return;

    log_file = fopen (LOGFILE, "a");

    if (log_file)
	setbuf (log_file, NULL);
}

/* Open the debug log */
void open_debug ()
{
    if (debug_file)
	return;

    debug_file = fopen ("debug.log", "a");

    if (debug_file)
	setbuf (debug_file, NULL);
}

/* Log something */
void log (const char *fmt,...)
{
    va_list args;
    time_t t;
    struct tm tm;
    char buf[32];

    va_start (args, fmt);
    time (&t);
    tm = *localtime (&t);
    strftime (buf, sizeof (buf) - 1, "[%d/%m %H:%M:%S] ", &tm);

    if (!log_file)
	open_log ();

    if (log_file)
    {
	fputs (buf, log_file);
	vfprintf (log_file, fmt, args);
	fputc ('\n', log_file);
	fflush (log_file);
    }

    if (runflags & RUN_LIVE)
    {
	fputs (buf, stderr);
	vfprintf (stderr, fmt, args);
	fputc ('\n', stderr);
    }
}

/* Write to our debug log */
void debug (const char *fmt,...)
{
    va_list args;
    time_t t;
    struct tm tm;
    char buf[32];

    va_start (args, fmt);
    time (&t);
    tm = *localtime (&t);
    strftime (buf, sizeof (buf) - 1, "[%d/%m %H:%M:%S] ", &tm);

    if (!debug_file)
	open_debug ();

    if (debug_file)
    {
	fputs (buf, debug_file);
	vfprintf (debug_file, fmt, args);
	fputc ('\n', debug_file);
	fflush (debug_file);
    }

    if (runflags & RUN_LIVE)
    {
	fputs (buf, stderr);
	vfprintf (stderr, fmt, args);
	fputc ('\n', stderr);
    }
}

/* Close the logfile */
void close_log ()
{
    if (!log_file)
	return;

    fclose (log_file);

    log_file = NULL;
}

/* Close the debug log */
void close_debug ()
{
    if (!debug_file)
	return;

    fclose (debug_file);

    debug_file = NULL;
}

/* Fatal error: Log it, globop it if we're still connected, then die. */
void fatal (const char *fmt,...)
{
    va_list args;
    char buf[BUFSIZE];

    va_start (args, fmt);    
    vsnprintf (buf, sizeof (buf), fmt, args);

    log ("Fatal error: %s", buf);

    printf ("\n\nFatal error: %s\n", buf);

    if (servsock > 0)
	globops ("Fatal Error: %s", buf);

    exit (1);
}

/* Send a command to the server.  The two forms here are like
   printf()/vprintf() and friends.
 */
void send_cmd (const char *source, const char *fmt,...)
{
    va_list args;
   
    va_start (args, fmt);
    vsend_cmd (source, fmt, args);
    va_end (args);
}

void vsend_cmd (const char *source, const char *fmt, va_list args)
{
    char buf[2048];
    
    vsnprintf (buf, sizeof (buf), fmt, args);

    if (source)
    {
	sockprintf (servsock, ":%s %s\r\n", source, buf);
	if (debuglevel == 1 || debuglevel == 3)
	    debug ("<-- :%s %s", source, buf);
    }
    else
    {
	sockprintf (servsock, "%s\r\n", buf);
	if (debuglevel == 1 || debuglevel == 3)
	    debug ("<-- %s", buf);
    }
}

/* Send a globops */
void globops (const char *sender, const char *fmt,...)
{
    va_list args;
    char buf[2048];
     
    va_start (args, fmt);
#ifdef IRCUP9
    /* IRCu doesn't have GlobOps. We'll use wallops. */
    snprintf (buf, sizeof (buf), "WALLOPS :%s", fmt);
#else
    snprintf (buf, sizeof (buf), "%s :%s", me.token ? "]" : "GLOBOPS", fmt);
#endif

    vsend_cmd (sender ? sender : me.name, buf, args);
    va_end (args);
}

/* PRIVMSG our snooping channel */
void snoop (const char *source, const char *fmt,...) 
{
    va_list args;
    char buf[2048];
    time_t t;
    struct tm tm;
    char timebuf[256];

    if (findchan (snoopchan))
    {
	time (&t);
	tm = *localtime (&t);
	strftime (timebuf, sizeof (timebuf) - 1, "[%H:%M:%S]", &tm);
	va_start (args, fmt);
	snprintf (buf, sizeof (buf), "%s %s :%s %s", me.token ? "!" :
	    "PRIVMSG", snoopchan, timebuf, fmt);
	vsend_cmd (source, buf, args);
	va_end (args);
    }
}

/* Send a notice to a user */
void notice (const char *source, const char *dest, const char *fmt,...)
{
    va_list args;
    char buf[2048];

    va_start (args, fmt);
 
    snprintf (buf, sizeof (buf), "%s %s :%s", me.token ? "B" : "NOTICE",
	dest, fmt);
    vsend_cmd (source, buf, args);
    va_end (args);
}

/* Reply to a CTCP command. */
void ctcpreply (const char *source, const char *dest, const char *fmt,...)
{
    va_list args;
    char buf[2048];

    va_start (args, fmt);
    snprintf (buf, sizeof (buf), "%s %s :\001%s\001", me.token ? "B" :
	"NOTICE", dest, fmt);
    vsend_cmd (source, buf, args);
    va_end (args);
}

/* Send a global notice to all users. */
void noticeall (const char *fmt,...)
{
#ifndef UNREAL3
    char *tmp = strtok (tlds, " ");
#endif

    /* Is the GlobalNoticer on? Bail out if not.
       (We could easily use OperServ for this, but if people don't have
       the GlobalNoticer on, they probably use something else for notices.)
     */
    if (globalnoticer_on == FALSE)
	return;

#ifdef UNREAL3
    /* Unreal lets us notice $*, so we'll do that. */
    send_cmd (s_GlobalNoticer, "%s $* :%s", me.token ? "B" : "NOTICE", fmt);
#else
    
    /* Notice the first TLD */
    send_cmd (s_GlobalNoticer, "%s $*.%s :%s", me.token ? "B" : "NOTICE",
	tmp, fmt);
        
    /* Now run through the list and notice the remaining tlds.
       This will terminate when we run out of tlds..
     */
    for (;;)
    {
	tmp = strtok (NULL, " ");

	if (tmp)
	    send_cmd (s_GlobalNoticer, "%s $*.%s :%s", me.token ? "B" :
		"NOTICE", tmp, fmt);
	else
	    break;
    }
#endif /* UNREAL3 */
}       

/* Case insensitive version of strcmp */
int stricmp (const char *s1, const char *s2)
{
    register int c;

    while ((c = tolower (*s1)) == tolower (*s2))
    {
	if (c == 0)
	    return 0;
	s1++;
	s2++;
    }
    if (c < tolower (*s2))
	return -1; 
    return 1;
}

/* strscpy:  Copy at most len-1 characters from a string to a buffer, and
             add a null terminator after the last character copied.
 */
char *strscpy (char *d, const char *s, size_t len)
{
    char *d_orig = d;

    if (!len)
	return d;

    while (len-- && (*d++ = *s++))
	;
    *d = 0;
    return d_orig;
}

void *smalloc (long size)
{
    void *buf;

    buf = malloc (size);
    if (!buf)
        raise (SIGUSR1);
    return buf;
}  

void *scalloc (long elsize, long els)  
{
    void *buf = calloc (elsize, els);
    if (!buf)
	raise (SIGUSR1);
    return buf;
}

void *srealloc (void *oldptr, long newsize)
{   
    void *buf = realloc (oldptr, newsize);
    if (!buf)
        raise (SIGUSR1);
    return buf;
}

char *sstrdup (const char *s)
{   
    char *t = strdup (s);
    if (!t)
        raise (SIGUSR1);
    return t;
}

/* Hash function for commands */
Hash *get_hash (const char *whoami, const char *source, const char *cmd,
		Hash *hash_table)
{
    for (; hash_table->accept; hash_table++)
	if (match_wild_nocase (hash_table->accept, cmd))
	{
	    if ((hash_table->access == H_IRCOP && is_oper (source)) ||
		(hash_table->access == H_SRA && is_sra (source)) ||
		(hash_table->access == H_NONE))
		return hash_table;
	    else if ((hash_table->access == H_IRCOP && !is_oper (source)) ||
		     (hash_table->access == H_SRA && !is_sra (source)))
	    {
		notice (whoami, source, ERR_ACCESS_DENIED, cmd);
		return NULL;
	    }
	}

    notice (whoami, source, ERR_UNKNOWN_COMMAND, cmd, haveserv_on == TRUE ? "" :
	"MSG ", whoami, haveserv_on == TRUE ? "" : securitysetting == 1 ? "@" :
	"", haveserv_on == TRUE ? "" : securitysetting == 1 ? me.name : "");

    return NULL;
}

/* Subcommand hash - Similar to above, but no error if the command isn't
   found, the function that called this will handle that.
 */
Hash_SUB *get_sub_hash (const char *whoami, const char *source, const char *cmd,
			Hash_SUB *hash_table)
{
    for (; hash_table->accept; hash_table++)
	if (match_wild_nocase (hash_table->accept, cmd))
	{
	    if ((hash_table->access == H_IRCOP && is_oper (source)) ||
		(hash_table->access == H_SRA && is_sra (source)) ||
		(hash_table->access == H_NONE))
		return hash_table;
	    else if ((hash_table->access == H_IRCOP && !is_oper (source)) ||
		     (hash_table->access == H_SRA && !is_sra (source)))
	    {
		notice (whoami, source, ERR_ACCESS_DENIED, cmd);
		return NULL;
	    }
	}

    return NULL;
}

/* Hash function for help tables */
Hash_HELP *get_help_hash (const char *whoami, const char *source,
			  const char *cmd, Hash_HELP *hash_table)
{
    for (; hash_table->accept; hash_table++)
	if (match_wild_nocase (hash_table->accept, cmd))
	{
	    if ((hash_table->access == H_IRCOP && is_oper (source)) ||
		(hash_table->access == H_SRA && is_sra (source)) ||
		(hash_table->access == H_NONE))
		return hash_table;
	    else if ((hash_table->access == H_IRCOP && !is_oper (source)) ||
		     (hash_table->access == H_SRA && !is_sra (source)))
	    {
		notice (whoami, source, ERR_ACCESS_DENIED, cmd);
		return NULL;
	    }
	}

    notice (whoami, source, ERR_NO_HELP, cmd);
    return NULL;
}

/* Is the given nick an oper? */
int is_oper (const char *nick)  
{
    User *user = finduser (nick);

    return (user && (user->mode & UMODE_o));
}

/* Is the given nick an SRA? */
int is_sra (const char *nick)
{
    User *user = finduser (nick);

    return (user && (user->flags & UF_SRA));
}

/* Search case-insensitively for string s2 within string s1 */
char *stristr (char *s1, char *s2)
{ 
    register char *s = s1, *d = s2;

    while (*s1)
	if (tolower (*s1) == tolower (*d))
	{
	    s1++;
	    d++; 
	    if (*d == 0)
		return s;
	}
	else
	{
	    s = s1++;
	    d = s2;  
	}
    return NULL;
}

/* Find a wildcard match */
static int do_match_wild (const char *pattern, const char *str, int  docase)
{
    char c;
    const char *s;

    for (;;)
    {
	switch (c = *pattern++)
	{
	    case 0:
		if (!*str)
		    return 1;
		return 0;
	    case '?':
		if (!*str)
		    return 0;
		str++;
		break;
	    case '*':
		if (!*pattern)
		    return 1;
		s = str;
		while (*s)
		{
		    if ((docase ? (*s==*pattern) : (tolower (*s)==tolower
			(*pattern))) && do_match_wild (pattern, s, docase))
			return 1;
		    s++;
		}
		break;
	    default:
		if (docase ? (*str++ != c) : (tolower(*str++) != tolower(c)))
		    return 0;
		break;
	}
    }
}

/* Call do_match_wild case-sensitively */
int match_wild (const char *pattern, const char *str)
{
    return do_match_wild (pattern, str, 1);
}

/* Call do_match_wild case-insensitively */
int match_wild_nocase (const char *pattern, const char *str)
{
    return do_match_wild (pattern, str, 0);
}

/* strupper: Convert a string to upper case. */
char *strupper (char *s)
{
    char *t = s;
    while (*t)
	*t++ = toupper (*t);
    return s;
}

/* strlower: Convert a string to lower case. */
char *strlower (char *s)
{
    char *t = s;
    while (*t)
	*t++ = tolower (*t);
    return s;
}

/* Strip unwanted chars out of lines - Borrowed from GeoStats */
void strip (char *line)
{  
    char *c;

    if (line)
    {
	if ((c = strchr (line, '\n')))
	    *c = '\0';
	if ((c = strchr (line, '\r')))
	    *c = '\0';
	if ((c = strchr (line, '\1')))
	    *c = '\0';
    }
}

/* Is it one of ours? */
int is_one_of_mine (const char *nick)
{
    if ((operserv_on == TRUE && !stricmp (nick, s_OperServ)) ||
        (statserv_on == TRUE && !stricmp (nick, s_StatServ)) ||
	(globalnoticer_on == TRUE && !stricmp (nick, s_GlobalNoticer)) ||
	(spamserv_on == TRUE && !stricmp (nick, sp.nick)))
        return 1;
 
    return 0;
}

/* Return the date an event happened
   If fulldate is:
	0:	Sun, Jan 01, 2000
	1:	Sun, Jan 01, 2000, 00:00:00 GMT
	2:	00:00:00 GMT
	3:	01/01/2000, 00:00:00 GMT
	4:	Jan 01, 2000, 00:00:00 GMT
 */
char *get_time (time_t event, int fulldate)
{
    struct tm tm;
    static char timebuf[256];

    tm = *gmtime (&event);

    /* Date and time, or just date */
    if (fulldate == 1)
	strftime (timebuf, sizeof (timebuf), "%a, %b %d, %Y, %H:%M:%S GMT",
	    &tm);
    else if (fulldate == 2)
	strftime (timebuf, sizeof (timebuf), "%H:%M:%S GMT", &tm);
    else if (fulldate == 3)
	strftime (timebuf, sizeof (timebuf), "%d/%m/%Y %H:%M:%S GMT", &tm);
    else if (fulldate == 4)
	strftime (timebuf, sizeof (timebuf), "%b %d, %Y, %H:%M:%S GMT", &tm);
    else
	strftime (timebuf, sizeof (timebuf), "%a, %b %d, %Y", &tm);

    timebuf[sizeof(timebuf)-1] = 0;
    return timebuf;
}

/* Return the time with the given offset. */
char *zone_time (int offset)
{
    struct tm tm;
    static char timebuf[256];
    time_t event;

    event = time (NULL) + offset;

    tm = *gmtime (&event);

    strftime (timebuf, sizeof (timebuf), "%H:%M:%S, %b %d, %Y", &tm);

    timebuf[sizeof(timebuf)-1] = 0;
    return timebuf;
}

/* Return the time elapsed since an event */
char *time_ago (time_t event)
{
    static char ret[128];
    int years, weeks, days, hours, minutes, seconds;

    event = time (NULL) - event;
    years = weeks = days = hours = minutes = seconds = 0;

    while (event >= 60 * 60 * 24 * 365)
    {
	event -= 60 * 60 * 24 * 365;
	years++;
    }
    while (event >= 60 * 60 * 24 * 7)
    {
	event -= 60 * 60 * 24 * 7;
	weeks++;
    }
    while (event >= 60 * 60 * 24)
    {
	event -= 60 * 60 * 24;
	days++; 
    }
    while (event >= 60 * 60)
    {
	event -= 60 * 60;
	hours++;
    }
    while (event >= 60)
    {
	event -= 60;
	minutes++;
    }

    seconds = event;
     
    if (years)
	snprintf (ret, sizeof (ret),
		"%d year%s, %d week%s, %d day%s, %02d:%02d:%02d",
		years, years == 1 ? "" : "s",
		weeks, weeks == 1 ? "" : "s",
		days, days == 1 ? "" : "s",
		hours, minutes, seconds);
    else if (weeks)
	snprintf (ret, sizeof (ret), "%d week%s, %d day%s, %02d:%02d:%02d",
		weeks, weeks == 1 ? "" : "s",
		days, days == 1 ? "" : "s",
		hours, minutes, seconds);
    else if (days)
	snprintf (ret, sizeof (ret), "%d day%s, %02d:%02d:%02d",
		days, days == 1 ? "" : "s",  
		hours, minutes, seconds);
    else if (hours)
	snprintf (ret, sizeof (ret), "%d hour%s, %d minute%s, %d second%s",
		hours, hours == 1 ? "" : "s",
		minutes, minutes == 1 ? "" : "s",
		seconds, seconds == 1 ? "" : "s");
    else if (minutes)
	snprintf (ret, sizeof (ret), "%d minute%s, %d second%s",
		minutes, minutes == 1 ? "" : "s",
		seconds, seconds == 1 ? "" : "s");
    else
	snprintf (ret, sizeof (ret), "%d second%s",
		seconds, seconds == 1 ? "" : "s");

    return ret;
}

/* Duration: How long since an event happened.
   Depending on format, we return a time in one of two formats:
	1: #d#h#m#s
	2: # days, # hours, # minutes, # seconds
   Like dotime, we only do days, hours, minutes and seconds here.
 */
char *duration (time_t event, int format)
{
    static char ret[128];
    int days, hours, minutes, seconds;

    days = hours = minutes = seconds = 0;

    while (event >= 60 * 60 * 24)
    {
	event -= 60 * 60 * 24;
	days++; 
    }
    while (event >= 60 * 60)
    {
	event -= 60 * 60;
	hours++;
    }
    while (event >= 60)
    {
	event -= 60;
	minutes++;
    }

    seconds = event;

    /* This is pretty big, but it works.. */
    if (format == 1)
    {
	if (days && hours && minutes && seconds)
	    snprintf (ret, sizeof (ret),
		"%dd%dh%dm%ds", days, hours, minutes, seconds);
	else if (days && hours && minutes)
	    snprintf (ret, sizeof (ret),
		"%dd%dh%dm", days, hours, minutes);
	else if (days && hours)
	    snprintf (ret, sizeof (ret),
		"%dd%dh", days, hours);
	else if (days)
	    snprintf (ret, sizeof (ret),
		"%dd", days);
	else if (hours && minutes && seconds)
	    snprintf (ret, sizeof (ret),
		"%dh%dm%ds", hours, minutes, seconds);
	else if (hours && minutes)
	    snprintf (ret, sizeof (ret),
		"%dh%dm", hours, minutes);
	else if (hours)
	    snprintf (ret, sizeof (ret),
		"%dh", hours);
	else if (minutes && seconds)
	    snprintf (ret, sizeof (ret),
		"%dm%ds", minutes, seconds);
	else if (minutes)
	    snprintf (ret, sizeof (ret), 
		"%dm", minutes);
	else
	    snprintf (ret, sizeof (ret),
		"%ds", seconds);
    }

    if (format == 2)
    {
	if (days && hours && minutes && seconds)
	    snprintf (ret, sizeof (ret),
		"%d day%s, %d hour%s, %d minute%s, %d second%s", days, days ==
		1 ? "" : "s", hours, hours == 1 ? "" : "s", minutes, minutes
		== 1 ? "" : "s", seconds, seconds == 1 ? "" : "s");
	else if (days && hours && minutes)
	    snprintf (ret, sizeof (ret),
		"%d day%s, %d hour%s, %d minute%s", days, days == 1 ? "" : "s",
		hours, hours == 1 ? "" : "s", minutes, minutes == 1 ? "" : "s");
	else if (days && hours)
	    snprintf (ret, sizeof (ret),
		"%d day%s, %d hour%s", days, days == 1 ? "" : "s",
		hours, hours == 1 ? "" : "s");
	else if (days)
	    snprintf (ret, sizeof (ret),
		"%d day%s", days, days == 1 ? "" : "s");
	else if (hours && minutes && seconds)
	    snprintf (ret, sizeof (ret),
		"%d hour%s, %d minute%s, %d second%s", hours, hours == 1 ? ""
		: "s", minutes, minutes == 1 ? "" : "s", seconds, seconds
		== 1 ? "" : "s");
	else if (hours && minutes)
	    snprintf (ret, sizeof (ret),
		"%d hour%s, %d minute%s", hours, hours == 1 ? "" : "s",
		minutes, minutes == 1 ? "" : "s");
	else if (hours)
	    snprintf (ret, sizeof (ret),
		"%d hour%s", hours, hours == 1 ? "" : "s");
	else if (minutes && seconds)
	    snprintf (ret, sizeof (ret),
		"%d minute%s, %d second%s", minutes, minutes  == 1 ? "" : "s",
		seconds, seconds == 1 ? "" : "s");
	else if (minutes)
	    snprintf (ret, sizeof (ret), 
		"%d minute%s", minutes, minutes == 1 ? "" : "s");
	else
	    snprintf (ret, sizeof (ret),
		"%d second%s", seconds, seconds == 1 ? "" : "s");
    }

    return ret;
}

/* dotime:  Return the number of seconds corresponding to the given time
            string.  If the given string does not represent a valid time,
            return -1.
  
            A time string is either a plain integer (representing a number
            of seconds), or an integer followed by one of these characters:
            "s" (seconds), "m" (minutes), "h" (hours), or "d" (days).

   Borrowed from IRCServices, modified to accept stacking of times. -skold
 */
int dotime (const char *s)
{
    int amount, seconds = 0;

    amount = strtol (s, (char **)&s, 10);

    while (*s)
    {
	switch (*s++)
	{
	    case 's':
		seconds += amount;
		amount = strtol (s, (char **)&s, 10);
		break;
	    case 'm':
		seconds += amount * 60;
		amount = strtol (s, (char **)&s, 10);
		break;
	    case 'h':
		seconds += amount * 3600;
		amount = strtol (s, (char **)&s, 10);
		break;
	    case 'd':
		seconds += amount * 86400;
		amount = strtol (s, (char **)&s, 10);
		break;
	}
    }
    return seconds;
}

/* Send a NULL-terminated array of text as NOTICEs. */
void notice_list (const char *source, const char *dest, const char **text)
{
    while (*text)
    {
	if (**text)
	    notice (source, dest, *text);
	else
	    notice (source, dest, " ");
	text++;  
    }
}

/* Find an SRA by nick. Return NULL if sra could not be found. */
SRA *findsra (const char *nick)
{
    SRA *sra;

    if (!nick)  
	return NULL;

    for (sra = sralist; sra; sra = sra->next)
	if (!stricmp (nick, sra->nick))
	    return sra;

    return NULL;
}

/* Find a ULine by name. Return NULL if uline could not be found. */
ULine *finduline (const char *name)
{
    ULine *uline;

    if (!name)
	return NULL;

    for (uline = ulinelist; uline; uline = uline->next)
	if (!stricmp (name, uline->name))
	    return uline;

    return NULL;
}

/* Find an TLD. Return NULL if tld could not be found. */
TLD *findtld (const char *param)
{
    TLD *tld;

    if (!param)
	return NULL;

    for (tld = tldlist; tld; tld = tld->next)
	if (!stricmp (param, tld->name) || !stricmp (param, tld->desc))
	    return tld;

    return NULL;
}

/* Find a TimeZone. Return NULL if TimeZone could not be found. */
Zone *findzone (const char *param)
{
    Zone *zone;

    if (!param)
	return NULL;
 
    for (zone = zonelist; zone; zone = zone->next)
	if (!stricmp (param, zone->name))
	    return zone;
 
    return NULL;
}

/* Find a server for stats purposes. */
ServStat *findserv (const char *server)
{
    ServStat *serverstat;

    serverstat = statlist[SHASH(server)];

    while (serverstat && !match_wild_nocase (server, serverstat->name))
	serverstat = serverstat->next;

    return serverstat;
}

/* Is the given mask akilled? */
int is_akilled (const char *mask)
{   
    AutoKill *akill;
    int i;
        
    for (i = 0; i < 1024; i++)
	for (akill = akilllist[i]; akill; akill = akill->next)
	    if (match_wild_nocase (akill->mask, mask))
		return 1;
    return 0;
}

/* Given a user, return a mask that will most likely match any address the
   user will have from that location.  For IP addresses, wildcards the
   appropriate subnet mask (e.g. 35.1.1.1 -> 35.*; 128.2.1.1 -> 128.2.*);
   for named addresses, wildcards the leftmost part of the name unless the
   name only contains two parts.  If the username begins with a ~, delete
   it.  The returned character string is malloc'd and should be free'd
   when done with.

   Borrowed from IRCServices.
 */
char *create_mask (User *u)
{
    char *mask, *s, *end;

    end = mask = smalloc (strlen (u->user) + strlen (u->host) + 2);
    end += sprintf (end, "*@");
    if (strspn (u->host, "0123456789.") == strlen (u->host)
	&& (s = strchr(u->host, '.')) && (s = strchr(s+1, '.'))
	&& (s = strchr(s+1, '.'))  && (   !strchr(s+1, '.')))
    {
	s = sstrdup (u->host);
	*strrchr (s, '.') = 0;
	if (atoi (u->host) < 192)
	    *strrchr (s, '.') = 0;
	if (atoi (u->host) < 128)
	    *strrchr (s, '.') = 0;
	sprintf (end, "%s.*", s);
	free(s);
    }
    else
    {
	if ((s = strchr (u->host, '.')) && strchr (s+1, '.'))
	{
	    s = sstrdup (strchr (u->host, '.') -1);
	    *s = '*';
	}
	else
	    s = sstrdup (u->host);

	strcpy (end, s);
	free (s);
    }
    return mask;
}

/* Remove control codes from text. Borrowed from UnrealIRCd,
   modified to strip bold, underline and reverse.
 */
char *stripcodes (char *buffer)
{
    static char in[BUFSIZE], out[BUFSIZE];
    int i = 0, j = 0, hascomma = 0;

    /* Check if theres any codes in the text */
    if (!strchr (buffer, '\002') && !strchr (buffer, '\003') &&
	!strchr (buffer, '\026') && !strchr (buffer, '\037'))
	return buffer;

    bzero ((char *)out, sizeof (out));
    memcpy (in, buffer, BUFSIZE);

    while (in[i])
    {
	/* Remove codes */
	if (in[i] == '\002' || in[i] == '\003' ||
	    in[i] == '\026' || in[i] == '\037')
	{
	    hascomma = 0;
	    i++;

	    if (!isdigit (in[i]))
		continue;

	    while (isdigit (in[i]) || (isdigit (in[i - 1])
		   && in[i] == ',' && isdigit (in[i + 1])
		   && hascomma == 0))
	    {
		if (in[i] == ',' && hascomma == 1)
		    break;
		if (in[i] == ',' && hascomma == 0)
		    hascomma = 1;
		i++;
	    }
	    continue;
	}
	out[j] = in[i];
	i++;
	j++;
    }
    return out;
}

/* Is it midnight? - Borrowed from GeoStats. */
int is_midnight ()
{
    time_t current = time (NULL);
    struct tm *ltm = gmtime (&current);

    if (ltm->tm_hour == 0)
	return 1;

    return 0;
}

/* Drop the current channel list to disk in an html format. 
   We get the format from an html file already on the disk.
   Theres several 'tokens' in this file that we look for, and
   insert information in place of them. The rest of the html
   we leave alone.
 */
void dumpchans ()
{
    FILE *in = fopen (CHANNELS_IN, "r");
    FILE *out = fopen (CHANNELS_OUT, "w");
    char last[64], end[64], topic[66], dBuf[2048], ret[2048], topictmp[2048];
    char *tmp, links[512], linktmp[512], logoimg[256];
    Channel *c;
    struct c_userlist *u;
    int j, k, rowcolor = 0;

    if (logupdates_on == TRUE)
	log (RPL_SAVING, CHANNELS_OUT);

    if (!in)
    {
	log (ERR_CANT_OPEN, CHANNELS_IN);
	return;
    }

    if (!out)
    {
	log (ERR_CANT_OPEN, CHANNELS_OUT);
	return;
    }

    /* Set up the token replacements ahead of time. */
    sprintf (last, "%s", get_time (time (NULL), 1));
    sprintf (end, "%d user%s and %d channel%s", userstat,
	userstat == 1 ? "" : "s", countchans (0),
	countchans (0) == 1 ? "" : "s");

    if (strlen (htmllogo))
	sprintf (logoimg, "<img src=\"%s\">", htmllogo);
    else
	sprintf (logoimg, "Generated by OperStats v%s", version);

    while (fgets (dBuf, 2047, in))
    {
	if (strstr (dBuf, "$C") || strstr (dBuf, "$U") || strstr (dBuf, "$T") ||
	    strstr (dBuf, "$R") || strstr (dBuf, "$H"))
	{
	    for (j = 0; j < 1024; j++)
	    {
		for (c = chanlist[j]; c; c = c->next)
		{
		    if (!(c->mode & CMODE_s))
		    {
			*links = 0;
			*linktmp = 0;
			k = 0;

			strscpy (ret, dBuf, sizeof (ret));

			if (rowcolor == 1)
			{
			    rowcolor = 0;

			    tmp = replace (ret, "$R", "rowdark");
			    strscpy (ret, tmp, sizeof (ret));
			}
			else
			{
			    rowcolor = 1;
			    tmp = replace (ret, "$R", "rowlight");
			    strscpy (ret, tmp, sizeof (ret));
			}

			if (strlen (mirclink))
			{
			    /* mIRCLinks don't play nice with channels with
			       the # in them. So we have to strip it off.
			     */
			    char *chantmp = sstrdup (c->name);

			    tmp = strchr (chantmp, '#');

			    *tmp++ = 0;

			    snprintf (links, sizeof (links),
				"<a href=\"irc://%s/%s\"><img src=\"%s\" border=\"0\" alt=\"mIRC Chat\"></a>",
				mirclink, tmp, mirclinkimg);

			    free (chantmp);
			}

			if (javaclient_on == TRUE)
			{
			    if (strlen (links))
			    {
				snprintf (linktmp, sizeof (linktmp), "%s",
				    links);

				snprintf (links, sizeof (links),
				    "%s<a href=\"javascript:openChat('%s')\"><img src=\"%s\" border=\"0\" alt=\"Java Chat\"></a>",
				    linktmp, c->name, javalinkimg);
			    }
			    else
			    {
				snprintf (links, sizeof (links),
				    "<a href=\"javascript:openChat('%s')\"><img src=\"%s\" border=\"0\" alt=\"Java Chat\"></a>",
				    c->name, javalinkimg);
			    }
			}

			if (strlen (links))
			{
			    tmp = replace (ret, "$H", links);
			    strscpy (ret, tmp, sizeof (ret));
			}

			tmp = replace (ret, "$C", c->name);
			strscpy (ret, tmp, sizeof (ret));

			for (u = c->users; u; u = u->next)
			    k++;

			/* Count our clients for the JoinChan if needed */
			if (!stricmp (c->name, joinchan))
			{
			    if (operserv_on == TRUE)
				k++;
			    if (statserv_on == TRUE)
				k++;
			    if (globalnoticer_on == TRUE)
				k++;
			}

			tmp = replace (ret, "$U", itoa (k));
			strscpy (ret, tmp, sizeof (ret));

			/* First, strip control codes from the topic. */
			if (c->topic)
			    tmp = stripcodes (c->topic);
			else
			    tmp = "";

			/* Now we want to cut the topic down to 65 chars
			   so it doesn't screw up the table.
			 */
			snprintf (topic, sizeof (topic), "%s", tmp);

			tmp = topic;

			/* Certain things mess up the html, fix them.
			   We replace these certain things with html
			   equivilants like &nbsp and &lt. These don't
			   affect the visible length, so we don't have
			   to limit it to 50 again.
			 */
			tmp = replace (tmp, "<", "&lt;");
			strscpy (topictmp, tmp, sizeof (topictmp));
			tmp = topictmp;
			tmp = replace (tmp, ">", "&gt;");
			strscpy (topictmp, tmp, sizeof (topictmp));
			tmp = topictmp;
			tmp = replace (tmp, " ", "&nbsp;");
			strscpy (topictmp, tmp, sizeof (topictmp));
			tmp = topictmp;

			tmp = replace (ret, "$T", tmp);
			strscpy (ret, tmp, sizeof (ret));
			fprintf (out, "%s", ret);
		    }
		}
	    }
	    continue;
	}

	if (strstr (dBuf, "$L"))
	{
	    strscpy (ret, dBuf, sizeof (ret));
	    tmp = replace (ret, "$L", last);
	    strscpy (ret, tmp, sizeof (ret));
	    fprintf (out, "%s", ret);
	    continue;
	}

	if (strstr (dBuf, "$E"))
	{
	    strscpy (ret, dBuf, sizeof (ret));
	    tmp = replace (ret, "$E", end);
	    strscpy (ret, tmp, sizeof (ret));
	    fprintf (out, "%s", ret);
	    continue;
	}

	if (strstr (dBuf, "$I"))
	{
	    strscpy (ret, dBuf, sizeof (ret));
	    tmp = replace (ret, "$I", logoimg);
	    strscpy (ret, tmp, sizeof (ret));
	    fprintf (out, "%s", ret);
	    continue;
	}

	fprintf (out, "%s", dBuf);
    }

    fclose (in);
    fclose (out);
}

/* Drop tld TLD stats to disk in an html format. This is similar
   to the TLDMAP output, except in html, and it's only done once
   a day.
 */
void dumptlds ()
{
    FILE *in = fopen (TLDSTATS_IN, "r");
    FILE *out = fopen (TLDSTATS_OUT, "w");
    TLD *tld;
    char last[64], end[64], dBuf[2048], ret[2048], logoimg[256], *tmp, *htmp;
    int i = 0, j;

    if (logupdates_on == TRUE)
	log (RPL_SAVING, TLDSTATS_OUT);

    if (!in)
    {
	log (ERR_CANT_OPEN, TLDSTATS_IN);
	return;
    }

    if (!out)
    {
	log (ERR_CANT_OPEN, TLDSTATS_OUT);
	return;
    }

    /* Set up the token replacements ahead of time. */
    for (tld = tldlist; tld; tld = tld->next)
	if (tld->hits)
	    i++;

    sprintf (last, "%s", get_time (time (NULL), 1));
    sprintf (end, "%d hit%s in %d TLD%s", dailyhitcnt,
	dailyhitcnt == 1 ? "" : "s", i, i == 1 ? "" : "s");

    if (strlen (htmllogo))
	sprintf (logoimg, "<img src=\"%s\">", htmllogo);
    else
	sprintf (logoimg, "Generated by OperStats v%s", version);

    while (fgets (dBuf, 2047, in))
    {
	if (strstr (dBuf, "$P") || strstr (dBuf, "$R") ||
	    strstr (dBuf, "$N") || strstr (dBuf, "$T") ||
	    strstr (dBuf, "$C") || strstr (dBuf, "$S"))
	{
	    for (tld = tldlist; tld; tld = tld->next)
		if (tld->hits)
		{
		    i = (tld->hits * 100) / dailyhitcnt;
		    j = 100 - i;

		    if (tld->hits == 1)
			htmp = "";
		    else
			htmp = "s";

		    strscpy (ret, dBuf, sizeof (ret));
		    tmp = replace (ret, "$P", itoa (i));
		    strscpy (ret, tmp, sizeof (ret));
		    tmp = replace (ret, "$R", itoa (j));
		    strscpy (ret, tmp, sizeof (ret));
		    tmp = replace (ret, "$N", tld->desc);
		    strscpy (ret, tmp, sizeof (ret));
		    tmp = replace (ret, "$T", tld->name);
		    strscpy (ret, tmp, sizeof (ret));
		    tmp = replace (ret, "$C", itoa (tld->hits));
		    strscpy (ret, tmp, sizeof (ret));
		    tmp = replace (ret, "$S", htmp);
		    strscpy (ret, tmp, sizeof (ret));
		    fprintf (out, "%s", ret);
		}

	    if (tldnoreshits)
	    {
		i = (tldnoreshits * 100) / dailyhitcnt;
		j = 100 - i;

		if (tldnoreshits == 1)
		    htmp = "";
		else
		    htmp = "s";

		strscpy (ret, dBuf, sizeof (ret));
		tmp = replace (ret, "$P", itoa (i));
		strscpy (ret, tmp, sizeof (ret));
		tmp = replace (ret, "$R", itoa (j));
		strscpy (ret, tmp, sizeof (ret));
		tmp = replace (ret, "$N", "Non Resolving");
		strscpy (ret, tmp, sizeof (ret));
		tmp = replace (ret, "$T", "Unknown");
		strscpy (ret, tmp, sizeof (ret));
		tmp = replace (ret, "$C", itoa (tldnoreshits));
		strscpy (ret, tmp, sizeof (ret));
		tmp = replace (ret, "$S", htmp);
		strscpy (ret, tmp, sizeof (ret));
		fprintf (out, "%s", ret);
	    }
	    continue;
	}

	if (strstr (dBuf, "$L"))
	{
	    strscpy (ret, dBuf, sizeof (ret));
	    tmp = replace (ret, "$L", last);
	    strscpy (ret, tmp, sizeof (ret));
	    fprintf (out, "%s", ret);
	    continue;
	}

	if (strstr (dBuf, "$E"))
	{
	    strscpy (ret, dBuf, sizeof (ret));
	    tmp = replace (ret, "$E", end);
	    strscpy (ret, tmp, sizeof (ret));
	    fprintf (out, "%s", ret);
	    continue;
	}

	if (strstr (dBuf, "$I"))
	{
	    strscpy (ret, dBuf, sizeof (ret));
	    tmp = replace (ret, "$I", logoimg);
	    strscpy (ret, tmp, sizeof (ret));
	    fprintf (out, "%s", ret);
	    continue;
	}

	fprintf (out, "%s", dBuf);
    }

    fclose (in);
    fclose (out);
}

/* Drop the current channels, users, and servers to a javascript file. */
void dumpstats ()
{
    FILE *f;

    if (logupdates_on == TRUE)
	log (RPL_SAVING, STATSJS);

    f = fopen (STATSJS, "w");

    if (!f)
    {
	log (ERR_CANT_OPEN, STATSJS);
	return;
    }

    /* Write the current users */
    fprintf (f, "curusers = \"%d\";\n", userstat);
 
    /* Write the current channels */
    fprintf (f, "curchans = \"%d\";\n", countchans (0));

    /* Write the current servers */
    fprintf (f, "curservs = \"%d\";\n", servstat);

    /* Write the max users */
    fprintf (f, "maxusers = \"%d\";\n", maxusercnt);

    /* Write the max channels */
    fprintf (f, "maxchans = \"%d\";\n", maxchancnt);

    /* Write the max servers */
    fprintf (f, "maxservs = \"%d\";\n", maxservcnt);

    fclose (f);
}

/* Replace oldpiece with newpiece in string. */
char *replace (char *string, char *oldpiece, char *newpiece)
{
    int str_index, newstr_index, oldpiece_index, end;
    int new_len, old_len, cpy_len;
    char *c;
    static char newstring[2048];

    if ((c = (char *) strstr (string, oldpiece)) == NULL)
	return string;

    new_len = strlen (newpiece);
    old_len = strlen (oldpiece);
    end = strlen (string) - old_len;
    oldpiece_index = c - string;

    newstr_index = 0;
    str_index = 0;

    while (str_index <= end && c != NULL)
    {
	/* Copy characters from the left of matched pattern occurence */
	cpy_len = oldpiece_index - str_index;
	strncpy (newstring + newstr_index, string + str_index, cpy_len);
	newstr_index += cpy_len;
	str_index += cpy_len;

	/* Copy replacement characters instead of matched pattern */
	strcpy (newstring + newstr_index, newpiece);
	newstr_index += new_len;
	str_index += old_len;

	/* Check for another pattern match */
	if ((c = (char *) strstr (string + str_index, oldpiece)) != NULL)
	   oldpiece_index = c - string;
    }

    /* Copy remaining characters from the right of last matched pattern */
    strcpy (newstring + newstr_index, string + str_index);

    return newstring;
}

/* Reverse of atoi (); */
char *itoa (int num)
{
    static char ret[32];
    sprintf (ret, "%d", num);
    return ret;
}

/* Clear all SRAs, TLDs, and TimeZones. */
void flushlists (void)
{
    SRA *sra;
    TLD *tld;
    Zone *zone;
    ULine *uline;

    for (sra = sralist; sra; sra = sra->next)
    {
	free (sra->nick);
	free (sra->pass);

	if (sra->prev)
	    sra->prev->next = sra->next;
	else
	    sralist = sra->next;

	if (sra->next)
	    sra->next->prev = sra->prev;
    }

    free (sra);

    for (tld = tldlist; tld; tld = tld->next)
    {
	free (tld->name);
	free (tld->desc);

	if (tld->prev)
	    tld->prev->next = tld->next;
	else
	    tldlist = tld->next;

	if (tld->next)
	    tld->next->prev = tld->prev;
    }

    free (tld);

    for (zone = zonelist; zone; zone = zone->next)
    {
	free (zone->name);
	free (zone->noffset);
	free (zone->desc);

	if (zone->prev)
	    zone->prev->next = zone->next;
	else
	    zonelist = zone->next;

	if (zone->next)
	    zone->next->prev = zone->prev;
    }

    free (zone);

    for (uline = ulinelist; uline; uline = uline->next)
    {
	free (uline->name);

	if (uline->prev)
	    uline->prev->next = uline->next;
	else
	    ulinelist = uline->next;

	if (uline->next)
	    uline->next->prev = uline->prev;
    }

    free (uline);
}

/* Return the number of IRCOps on the network */
int countops (void)
{
    int i, found = 0;
    User *u;

    for (i = 0; i < 1024; i++)
	for (u = userlist[i]; u; u = u->next)
	    if (u->mode & UMODE_o)
		found++;

    /* Whups, don't forget our own clients! */
    if (strchr (os.mode, 'o'))
	found++;

    if (strchr (ss.mode, 'o'))
	found++;

    if (strchr (gn.mode, 'o'))
	found++;

    return found;
}

/* SpamServ routine. */
void spamserv (const char *source, char *buf)
{
    char *cmd, *s;
    char orig[BUFSIZE];

    /* Make a copy of the original message. */
    strscpy (orig, buf, sizeof (orig));

    cmd = strtok (buf, " ");

    if (!cmd)
	return;

    if (!stricmp (cmd, "\1PING"))
    {
	if (!(s = strtok (NULL, "")))
	    s = "0";
	strip (s);
	ctcpreply (sp.nick, source, "PING %s", s);
	return;
    }
    else if (!stricmp (cmd, "\1VERSION\1"))
    {
	ctcpreply (sp.nick, source, "VERSION mIRC32 v5.9 K.Mardam-Bey");
	return;
    }

    /* A CTCP we don't recognize.. Ignore it. */
    else if (*cmd == '\1')
	return;

    /* Opers seem to find it funny when SpamServ globops their messages.
       NO FUN FOR YOU.
     */
    else if (is_oper (source))
	return;

    /* If we get here, we'll globop it. Theres no reason I can think of
       to message someone whos not in any channels and never messaged
       you first unless you're spamming them for something. So we consider
       it spam and send a globops about possible spam.
     */
    globops (operserv_on == TRUE ? s_OperServ : s_StatServ, RPL_POSSIBLE_SPAM,
	source, orig);
}

/* Change struct timeval into integer milliseconds. */
int tv2ms (struct timeval *tv)
{
    return (tv->tv_sec * 1000) + (int) (tv->tv_usec / 1000);
}

/* Send 'ERR_MORE_INFO' in the needed format. */
void errmoreinfo (const char *whoami, const char *source, char *text)
{
    notice (whoami, source, ERR_MORE_INFO, haveserv_on == TRUE ? "" : "MSG ",
	whoami, haveserv_on == TRUE ? "" : securitysetting == 1 ? "@" : "",
	haveserv_on == TRUE ? "" : securitysetting == 1 ? me.name : "", text);
}
